#!/usr/bin/python

# phi4npt.py: Perturbation theory for n-component phi^4 field theory
# (requires Python 2.6 or higher)

# Roland Bauerschmidt <roland@bauerschmidt.ca>
# David C. Brydges <db5d@math.ubc.ca>
# Gordon Slade <slade@math.ubc.ca>

# Changelog:
# - Correction of sign in [rg-loc, (1.52)]
# - For debugging print action of Loc on monomials
# - Correction of Loc(phi(nabla phi))

from wick import *

# number of components
N = 1
# N = 3
# N = 5


# coupling constants
half   = Fraction(1,2)
g      = Coupling('g')
nu     = Coupling('\\nu')
nuplus = Coupling('\\nu_+')
y      = Coupling('y')
z      = Coupling('z')

taux   = 0
tauy   = 0
taulx  = 0
tauly  = 0
taunx  = 0
tauny  = 0
phix = {}
phiy = {}
lphix = {}
lphiy = {}
nphix = {}
nphiy = {}

for n in range(0,N):

	# a complex Boson and Fermion field (\phi, \bar\phi, \psi, \psi\bar)
	phix[n]   = Field('\\phi%d' % n,               'x', fermion=0, dimension=1)
	phiy[n]   = Field('\\phi%d' % n,               'y', fermion=0, dimension=1)

	# Laplace operator applied to the Fields
	lphix[n]  = Field('-\\Delta \\phi%d' % n,      'x', fermion=0, dimension=3)
	lphiy[n]  = Field('-\\Delta \\phi%d' % n,      'y', fermion=0, dimension=3)

	# Nabla operator applied to the Fields
	nphix[n]  = Field('\\nabla \\phi%d' % n,       'x', fermion=0, dimension=2)
	nphiy[n]  = Field('\\nabla \\phi%d' % n,       'y', fermion=0, dimension=2)

	# quadratic and quartic supersymmetric combinations
	if n == 0:
		taux   = phix[n]*phix[n]*half
		tauy   = phiy[n]*phiy[n]*half
		taulx  = phix[n]*lphix[n]*half
		tauly  = phiy[n]*lphiy[n]*half
		taunx  = nphix[n]*nphix[n]*half
		tauny  = nphiy[n]*nphiy[n]*half
	else:
		taux   += phix[n]*phix[n]*half
		tauy   += phiy[n]*phiy[n]*half
		taulx  += phix[n]*lphix[n]*half
		tauly  += phiy[n]*lphiy[n]*half
		taunx  += nphix[n]*nphix[n]*half
		tauny  += nphiy[n]*nphiy[n]*half

tau2x  = taux*taux
tau2y  = tauy*tauy

# This function depends on the fields and a string that labels
# the covariance. Subsequently, new functions, C, W, and Wplus, are
# defined that are the same as __C, but have the string that label
# the covariance set to appropriate values
def __C(f1, f2, C='C'):

	l = []
	for n in range(0,N):
		l += [
			( phix[n],   phix[n], C+'(0)'),
			( phiy[n],   phiy[n], C+'(0)'),
			( phix[n],   phiy[n], C+'(x-y)'),
			(lphix[n],   phix[n], '-\\Delta '+C+'(0)'),
			(lphiy[n],   phiy[n], '-\\Delta '+C+'(0)'),
			(lphix[n],   phiy[n], '-\\Delta '+C+'(x-y)'),
			(lphiy[n],   phix[n], '-\\Delta '+C+'(x-y)'),
			( phix[n],  lphix[n], '-\\Delta '+C+'(0)'),
			( phiy[n],  lphiy[n], '-\\Delta '+C+'(0)'),
			( phix[n],  lphiy[n], '-\\Delta '+C+'(x-y)'),
			( phiy[n],  lphix[n], '-\\Delta '+C+'(x-y)'),
			(lphix[n],  lphix[n], '\\Delta^2 '+C+'(0)'),
			(lphiy[n],  lphiy[n], '\\Delta^2 '+C+'(0)'),
			(lphix[n],  lphiy[n], '\\Delta^2 '+C+'(x-y)'),
			(lphiy[n],  lphix[n], '\\Delta^2 '+C+'(x-y)'),
			(nphix[n],  nphix[n], '\\nabla\\nabla^* '+C+'(0)'),
			(nphiy[n],  nphiy[n], '\\nabla\\nabla^* '+C+'(0)'),
			(nphix[n],  nphiy[n], '\\nabla\\nabla^* '+C+'(x-y)'),
			(nphiy[n],  nphix[n], '\\nabla\\nabla^* '+C+'(x-y)'),
			(lphix[n],  nphix[n], '-\\Delta\\nabla^* '+C+'(0)'),
			(lphiy[n],  nphiy[n], '-\\Delta\\nabla '+C+'(0)'),
			(lphix[n],  nphiy[n], '-\\Delta\\nabla^* '+C+'(x-y)'),
			(lphiy[n],  nphix[n], '-\\Delta\\nabla '+C+'(x-y)'),
			( phix[n],  nphiy[n], '\\nabla^* '+C+'(x-y)'),
			( phiy[n],  nphix[n], '\\nabla '+C+'(x-y)'),
		]

	for rf1, rf2, rC in l:
		if str(f1) == str(rf1) and str(f2) == str(rf2):
			# this may not be needed anymore
			if rC[0] == '-':
				rC = rC[1:]
				sign = -1
			else:
				sign = 1

			return Coupling(rC) * sign


# definition of covariances C, W, and W_+
C     = (lambda f1, f2: __C(f1,f2,'C'))
W     = (lambda f1, f2: __C(f1,f2,'W'))
Wplus = (lambda f1, f2: __C(f1,f2,'W_+'))

# definition of F as in rg-pt
def F(C, U, V):
	UE = U.E(C, -1)
	VE = V.E(C, -1)
	return (UE*VE).E(C, 1) - U*V

# definition of Loc as in rg-loc
def Loc(i, pos, dim=4):

	# i is a monomial: use definition from rg-loc
	if isinstance(i, Monomial):

		if i.dimension() > dim:
			return Monomial(0)

		fields = []
		for f in i.fields:
			fields.append(Field(f.name, pos=pos, fermion=f.fermion, dimension=f.dimension))
		m = Monomial(fields)
		m.simplify()
		p = Polynomial(m)

		if i.degree() == 2:

			f1 = i.fields[0]
			f2 = i.fields[1]

                        # when f1 or f2 contain derivatives these
                        # definitions create names not in the tables
                        # of covariances

			f1pos = Field(f1.name, pos, fermion=f1.fermion, dimension=f1.dimension)
			nf1pos = Field('\\nabla ' + f1.name, pos, fermion=f1.fermion, dimension=f1.dimension+1)
			nf2pos = Field('\\nabla ' + f2.name, pos, fermion=f2.fermion, dimension=f2.dimension+1)
			lf1pos = Field('-\\Delta ' + f1.name, pos, fermion=f1.fermion, dimension=f1.dimension+2)
			lf2pos = Field('-\\Delta ' + f2.name, pos, fermion=f2.fermion, dimension=f2.dimension+2)

			# this is only because of fermions below and should be fixed
			if pos != 'x':
				raise 'Cannot localize to '+pos

			# (1.49) in rg-loc
			if f1.pos == f2.pos:
				if f1.pos != pos:
					dxy = Coupling('(' + pos + '_1-' + f1.pos +'_1)^2')
					# sigma = -taulx + taunx -- but we need this for fermions and bosons separately
					if f1.fermion:
						# needs to be generalized as for bosons below
						sigma_psi = (psibx*lpsix + lpsibx*psix)*(-half) + npsibx*npsix
						p += sigma_psi * dxy
                                                # chdb return p 
					else:
						#sigma_phi = (phibx*lphix + lphibx*phix)*(-half) + nphibx*nphix
						sigma_phi = -(f1pos*lf1pos) + nf1pos*nf1pos
						p += sigma_phi * dxy
                                                # chdb: add next line
                                                # else == included in next else
                                                return p
                                # chdb: # else == included in next else
                                else:
                                        return p

			# (1.52) in rg-loc
                        if f1.dimension+f2.dimension+2 <= 4:
				if f1.pos < f2.pos:
					dxy = Coupling('(' + f1.pos + '_1-' + f2.pos +'_1)^2')
				else:
					dxy = Coupling('(' + f2.pos + '_1-' + f1.pos +'_1)^2')
				if f1.pos == pos:
					# p += f1 * lf2pos * dxy * half
                                        p += - f1 * lf2pos * dxy * half    # correction to line above
				else:
					# p += lf1pos * f2 * dxy * half
                                      	p += - lf1pos * f2 * dxy * half    # correction to line above

                        # Loc for (nabla phi) phi
                        # as in greek.tex Section~\ref{sec:nablaphiphi}.
                        # adds correction to p for moving field not at x to x
                        
                        elif f1.dimension+f2.dimension == 3:
                                if i.fields[0].pos == i.fields[1].pos:
                                        return p
                                # case f1.pos = pos; handle other case by switching
                                if f1.pos == pos:
                       			fnew1 = i.fields[0]
			                fnew2 = i.fields[1]
                                else:
                                        fnew1 = i.fields[1]
			                fnew2 = i.fields[0]
                                dxy = '(\\nabla_x^t)^{-1}'
                                # nabla phi_x \phi_y -> (nabla_x^t)^{-1}(nabla phi)_x^2
			 	if fnew2.dimension == 1:
                                        nfnew2pos = Field('\\nabla ' + fnew2.name, pos, fermion=fnew2.fermion, dimension=fnew2.dimension+1)
			 		p += fnew1*nfnew2pos*dxy
                                # nabla phi_y phi_x  -> - (nabla_x^t)^{-1} (-Delta phi_x) phi_x
			 	else:
                                        lfnew2pos = Field(fnew2.name.replace('\\nabla ','-\\Delta '), pos, fermion=fnew2.fermion, dimension=fnew2.dimension)
			 	        p += - fnew1*lfnew2pos*dxy
		return p

	# i is a polynomial: localize each monomial separately
	elif isinstance(i, Polynomial):

		p = Polynomial()
		for g, m in i.monoms:
			p += Loc(m, pos, dim=dim)*g
		return p

	else:
		raise TypeError("%s must be Polynomial or Monomial." % i)

# for debug purposes: print Loc(m) for all monomials m in mons
if 0:
	print('The following definition of Loc is used below:\n')
	mons = [# phix[0]*phiy[0],
			# phix[0]*nphiy[0],
                	# nphix[0]*phiy[0],
			      # phix[0]*lphiy[0],
                              # lphix[0]*phiy[0],
                                 # phiy[0]*phiy[0],
                                  # taux
        ]
	for m in mons:
		print('----\nLoc(%s) =\n%s\n' % (m, Loc(m,'x')))

print('\n\nThe result for P_x is as follows:\n')

# Finally, calculate "P" as in rg-pt.
Vx  = tau2x * g + taux * nu     + taulx * z + taunx * y
Vy  = tau2y * g + tauy * nu     + tauly * z + tauny * y
EVx = tau2x * g + taux * nuplus + taulx * z + taunx * y
EVy = tau2y * g + tauy * nuplus + tauly * z + tauny * y

# 1st order
# Px = Vx.E(C)
# 2nd order
Px = (Loc(F(Wplus, EVx, EVy),'x') - Loc(F(W, Vx, Vy),'x').E(C))*half

# change to "tau"-basis
taux_n  = Field('\\tau', 'x')
taulx_n = Field('\\tau_\\Delta', 'x')
taunx_n = Field('\\tau_{\\nabla\\nabla}', 'x')
tau_coupling  = Px.proj(phix[0]*phix[0])*2
tau2_coupling = Px.proj(phix[0]*phix[0]*phix[0]*phix[0])*4
taul_coupling = Px.proj(phix[0]*lphix[0])*2
taun_coupling = Px.proj(nphix[0]*nphix[0])*2
Px = Px - taux  * tau_coupling  + taux_n    * tau_coupling
Px = Px - tau2x * tau2_coupling + taux_n**2 * tau2_coupling
Px = Px - taulx * taul_coupling + taulx_n   * taul_coupling
Px = Px - taunx * taun_coupling + taunx_n   * taun_coupling
Px.simplify()

print Px
